-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------

-- Basically a copy of the list I made for NRT.
local ignoredItems = {
	[30316] = true, -- Devastation
	[30317] = true, -- Cosmic Infuser
	[30312] = true, -- Infinity Blade
	[30318] = true, -- Netherstrand Longbow
	[30313] = true, -- Staff of Disintegration
	[30311] = true, -- Warp Slicer
	[30320] = true, -- Bundle of Nether Spikes
	[30314] = true, -- Phaseshift Bulwark
	[30319] = true, -- Nether Spikes
	[29434] = true, -- Badge of Justice
	[22450] = true, -- Void Crystal
	[20725] = true, -- Nexus Crystal

	[32897] = true, -- Mark of the Illidari
	[32428] = true, -- Heart of Darkness
	[34664] = true, -- Sunmote

	[32230] = true, -- Shadowsong Amethyst
	[32249] = true, -- Seaspray Emerald
	[32231] = true, -- Pyrestone
	[32229] = true, -- Lionseye
	[32228] = true, -- Empyrean Sapphire
	[32227] = true, -- Crimson Spinel
}

local _G = getfenv(0)

local deformat = AceLibrary("Deformat-2.0")
local tablet = AceLibrary("Tablet-2.0")

local db = nil

local guildMemberList = nil
local hexColors = {
	WTF = "|cffa0a0a0"
}
for k, v in pairs(RAID_CLASS_COLORS) do
	hexColors[k] = "|cff" .. string.format("%02x%02x%02x", v.r * 255, v.g * 255, v.b * 255)
end
local coloredName = setmetatable({}, {__index =
	function(self, key)
		if type(key) == "nil" then return nil end
		local class = guildMemberList and guildMemberList[key] or select(2, UnitClass(key)) or "WTF"
		self[key] = hexColors[class]  .. key .. "|r"
		return self[key]
	end
})

local knownClients = {}
local takingBids = nil
local tmpBidTable = {}
local lastDKPUpdate = nil
local hostOffline = nil
local activeModule = nil
local whisperHelp = {}
local whisperHandlers = {}
local recentWhispers = {}
local bidders = {}
local playerName = UnitName("player")

local function dkpSorter(a, b)
	local aDKP = activeModule:GetDKP(a) or 0
	local bDKP = activeModule:GetDKP(b) or 0
	return aDKP > bDKP and true or false
end

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = setmetatable({}, {__index =
	function(self, key)
		self[key] = key
		return key
	end
})

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

BananaDKP = AceLibrary("AceAddon-2.0"):new(
	"AceEvent-2.0",
	"AceConsole-2.0",
	"AceDB-2.0",
	"FuBarPlugin-2.0",
	"AceModuleCore-2.0",
	"AceHook-2.1"
)
local BananaDKP = BananaDKP

BananaDKP:SetModuleMixins("AceEvent-2.0")

local options = {
	type = "group",
	handler = BananaDKP,
	args = {
		dkpType = {
			type = "group",
			name = L["Type"],
			desc = L["Set what DKP system to use. Everyone must use the same system."],
			args = {},
			order = 101,
			disabled = "~IsActive",
		},
		dkpTypeOptions = {
			type = "group",
			name = L["Type options"],
			desc = L["Options for the current DKP type."],
			args = {},
			order = 102,
			disabled = function()
				if not IsRaidLeader() and not IsRaidOfficer() then return true end
				return not BananaDKP:IsActive()
			end,
		},
		role = {
			type = "group",
			name = L["Role"],
			desc = L["Sets your role, either client or server."],
			args = {
				client = {
					type = "toggle",
					name = L["Client"],
					desc = L["Sets this BananaDKP to act as a client."],
					get = function() return db.mode == 2 end,
					set = function(v)
						db.mode = v and 2 or 1
						BananaDKP:TriggerEvent("BananaDKP_ModeChanged")
					end,
				},
				server = {
					type = "toggle",
					name = L["Server"],
					desc = L["Sets this BananaDKP to act as a server."],
					get = function() return db.mode == 1 end,
					set = function(v)
						db.mode = v and 1 or 2
						BananaDKP:TriggerEvent("BananaDKP_ModeChanged")
					end,
				},
				host = {
					type = "text",
					name = L["Host"],
					desc = L["If acting as a client, set the host (player name) that is running the server here."],
					get = function() return db.host end,
					set = function(v)
						db.host = v
						BananaDKP:TriggerEvent("BananaDKP_HostChanged")
					end,
					usage = "<player>",
					disabled = function() return db.mode ~= 2 end,
				},
			},
			disabled = "~IsActive",
		},
		sitouts = {
			type = "group",
			name = L["Sitouts"],
			desc = L["List of players sitting out this raid who should still receive DKP. Click on a name to remove him from the list."],
			disabled = function() return type(currentLeader) ~= "boolean" or not currentLeader end,
			args = {
				add = {
					type = "text",
					name = L["Add player"],
					desc = L["Adds a player to the sitout list."],
					get = false,
					usage = "<player name>",
					set = "AddSitoutPlayer",
					order = 1,
				},
				clear = {
					type = "execute",
					name = L["Clear list"],
					desc = L["Clear the sitout list."],
					func = "ClearSitoutList",
					order = 2,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 3,
				},
			},
			order = 200,
			disabled = function()
				if not UnitInRaid("player") then return true end
				if not IsRaidLeader() and not IsRaidOfficer() then return true end
				return not BananaDKP:IsActive()
			end,
		},
	},
}

-----------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------

-- Meh, can't set this in OnInitialize, since OnEmbedInitialize is called for
-- FBP before our OnInit, and FBP requires the property to be set then.
BananaDKP.hasIcon = "Interface\\Icons\\Ability_Hunter_AspectOfTheMonkey"

function BananaDKP:OnInitialize()
	local revision = tonumber((string.match("$Revision: 66235 $", "^.-(%d+).-$"))) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.BANANADKP_VERSION = self.version
	_G.BANANADKP_REVISION = self.revision

	self:RegisterDB("BananaDKPDB")
	self:RegisterDefaults("profile", {
		sitouts = {}, -- Remember this across sessions.
		dkpType = "Symphony", -- The DKP module to use.
		mode = 2, -- 1 = server, 2 = client
		host = nil, -- No host set by default.
		-- playerId = {}, -- Key: Name, Value: ID. Not using this, we're just
		-- using the player names for keys now - this should be fixed, or maybe
		-- an alias table could be added, for player alts that should have the
		-- same DKP.
	})

	self.OnMenuRequest = options
	self:RegisterChatCommand("/bdkp", options, "BANANADKP")

	db = self.db.profile
end

function BananaDKP:OnEnable(first)
	local currentModule = db.dkpType
	for name, mod in self:IterateModules() do
		if name == currentModule then
			self:SetActiveDKPModule(name)
			break
		end
	end

	if not activeModule then
		self:Print(L["Error: No DKP module set, please reset your saved variables."])
		return
	end

	-- For hiding DKP whispers.
	self:Hook("ChatFrame_MessageEventHandler", "WhisperHandler", true)

	self:RegisterEvent("BananaDKP_DKPUpdated")
	self:RegisterEvent("CHAT_MSG_LOOT")
	self:RegisterEvent("GUILD_ROSTER_UPDATE")

	self:RebuildSitoutList()
end

function BananaDKP:OnDisable()
	for name, module in self:IterateModules() do
		self:ToggleModuleActive(module, false)
	end
end

-----------------------------------------------------------------------
-- Events
-----------------------------------------------------------------------

-- Someone thinks we should be synchronizing.
function BananaDKP:BananaDKP_DKPUpdated(remoteClient)
	if remoteClient then
		local found = nil
		for i, v in ipairs(knownClients) do
			if v:find(remoteClient) then
				found = true
				break
			end
		end
		if not found then
			table.insert(knownClients, coloredName[remoteClient])
		end
	end
	lastDKPUpdate = time()
	self:UpdateDisplay()
end

function BananaDKP:CHAT_MSG_LOOT(msg)
	if not UnitInRaid("player") then return end

	local player, item = deformat(msg, LOOT_ITEM)
	if not player then
		item = deformat(msg, LOOT_ITEM_SELF)
		if item then
			player = playerName
		end
	end

	if type(item) == "string" and type(player) == "string" then
		local itemName, itemLink, itemRarity = GetItemInfo(item)
		-- Find the item ID to make sure it's not a ignored item
		local itemId = select(3, itemLink:find("item:(%d+):"))
		if not itemId then return end
		itemId = tonumber(itemId:trim())
		if type(itemId) ~= "number" or ignoredItems[itemId] then return end

		-- Unconditionally notify modules regardless of itemRarity. Of course,
		-- most times modules will only care if itemRarity is higher than 3
		-- (epic or higher), but you never know.
		if type(itemName) == "string" and type(itemLink) == "string" and type(itemRarity) == "number" then
			activeModule:ItemLooted(player, itemLink, itemRarity)
		end
	end
end

function BananaDKP:GUILD_ROSTER_UPDATE()
	if not guildMemberList then
		self:UnregisterEvent("GUILD_ROSTER_UPDATE")
		guildMemberList = {}
		local offline = GetGuildRosterShowOffline()
		local selection = GetGuildRosterSelection()
		SetGuildRosterShowOffline(true)
		SetGuildRosterSelection(0)
		GetGuildRosterInfo(0)
		local numGuildMembers = GetNumGuildMembers()
		for i = 1, numGuildMembers do
			local name, _, _, _, _, _, _, _, _, _, class = GetGuildRosterInfo(i)
			guildMemberList[name] = class
		end
		SetGuildRosterShowOffline(offline)
		SetGuildRosterSelection(selection)
	end
end

-----------------------------------------------------------------------
-- Hooks
-----------------------------------------------------------------------

local function tablePrinter(t, recipient)
	for i, v in ipairs(t) do
		if type(v) == "string" then
			BananaDKP:Chat(L["BDKP: %s."]:format(v), "WHISPER", nil, recipient)
		elseif type(v) == "table" then
			tablePrinter(v, recipient)
		end
	end
end

local function bidTimeout()
	for k in pairs(tmpBidTable) do tmpBidTable[k] = nil end
	for k in pairs(bidders) do table.insert(tmpBidTable, k) end
	if #tmpBidTable > 0 then
		table.sort(tmpBidTable, dkpSorter)
		local winner = tmpBidTable[1]
		local winnerDkp = activeModule:GetDKP(winner)
		if #tmpBidTable > 1 and winnerDkp == activeModule:GetDKP(tmpBidTable[2]) then
			local more = true
			while more do
				table.remove(tmpBidTable, #tmpBidTable)
				more = activeModule:GetDKP(tmpBidTable[#tmpBidTable]) ~= winnerDkp
			end
			BananaDKP:Chat(L["Tie: %s."]:format(table.concat(tmpBidTable, ", ")), "RAID")
		else
			BananaDKP:Chat(L["%s wins (%d dkp)!"]:format(winner, winnerDkp), "RAID")
		end
	end
	takingBids = nil
end

local function newBidder(bidder)
	-- Force a DKP update in case we're in client mode.
	local dkp = activeModule:GetDKP(bidder)
	bidders[bidder] = true
	BananaDKP:ScheduleEvent("BidTimeout", bidTimeout, 15)
end

local notSeenString = "^" .. _G.ERR_CHAT_PLAYER_NOT_FOUND_S:gsub("%%s", "(.-)") .. "$"
local offlineString = "^" .. _G.ERR_FRIEND_OFFLINE_S:gsub("%%s", "(.-)") .. "$"
function BananaDKP:WhisperHandler(event, ...)
	if db.mode == 2 and UnitInRaid(db.host) then return self.hooks["ChatFrame_MessageEventHandler"](event, ...) end
	if (event == "CHAT_MSG_WHISPER" or event == "CHAT_MSG_WHISPER_INFORM") and type(arg1) == "string" and type(arg2) == "string" then
		local text = arg1:lower():trim()
		local player = arg2
		if text:find("^bdkp") then
			if event == "CHAT_MSG_WHISPER_INFORM" then return end
			local t = recentWhispers[player]
			if not t or (GetTime() - t > 1) then
				recentWhispers[player] = GetTime()
				local command, rest = select(3, text:find("^b.-dkp (%w+)(.*)"))
				if command then
					command = command:gsub("%A", "")  -- So we don't disconnect if someone tries to send us item links.
				end
				if rest then
					rest = rest:gsub("%A", ""):gsub("^%l", string.upper)
					if rest:len() < 2 then rest = nil end
				end
				if command and type(whisperHandlers[command]) == "function" then
					local ret = whisperHandlers[command](event, rest or player)
					if type(ret) == "string" and ret:trim():len() > 3 then
						self:Chat(L["BDKP: %s."]:format(ret), "WHISPER", nil, player)
					elseif type(ret) == "table" then
						tablePrinter(ret, player)
					end
				else
					local input = command and command:gsub("^%l", string.upper) or player
					local dkp = activeModule:GetDKP(input)
					if type(dkp) == "number" then
						self:Chat(L["BDKP: DKP for %s: %s."]:format(input, tostring(dkp)), "WHISPER", nil, player)
					else
						self:Chat(L["BDKP: Unknown command or player not found."], "WHISPER", nil, player)
					end
				end
				return
			end
		elseif takingBids and text:find("^bid") then
			newBidder(player)
			return
		elseif text == "sitout" and event == "CHAT_MSG_WHISPER" then
			self:ProcessSitoutWhisper(player)
			return
		end
	elseif event == "CHAT_MSG_RAID" or event == "CHAT_MSG_RAID_LEADER" and type(arg2) == "string" and db.mode == 1 then
		if arg2 == playerName and arg1:find(L["^bid:"]) then
			takingBids = true
			for k in pairs(bidders) do bidders[k] = nil end
			self:ScheduleEvent("BidTimeout", bidTimeout, 15)
		elseif takingBids and arg1:lower():find(L["^bid"]) then
			newBidder(arg2)
		end
	elseif event == "CHAT_MSG_SYSTEM" then
		if #knownClients > 0 then
			local player = arg1:match(offlineString)
			if player then
				for i, v in ipairs(knownClients) do
					if v:find(player) then
						table.remove(knownClients, i)
						break
					end
				end
			end
		end
		if db.host and not hostOffline then
			local player = arg1:match(offlineString)
			if not player then player = arg1:match(notSeenString) end
			if player and db.host == player then
				self:Print(L["Your host (%s) is offline. Please click the BananaDKP icon when he is online to continue updating."]:format(player))
				hostOffline = true
				self:TriggerEvent("BananaDKP_HostChanged")
				return
			end
		end
	end
	return self.hooks["ChatFrame_MessageEventHandler"](event, ...)
end

-----------------------------------------------------------------------
-- Communication Handling
-----------------------------------------------------------------------

function BananaDKP:Chat(text, chatType, language, destination)
	if ChatThrottleLib then
		ChatThrottleLib:SendChatMessage("BULK", "BananaDKP", text, chatType, language, destination)
	else
		SendChatMessage(text, chatType, language, destination)
	end
end

-----------------------------------------------------------------------
-- FuBar Handling
-----------------------------------------------------------------------

function BananaDKP:OnTextUpdate()
	self:SetText(db.mode == 1 and L["Server"] or L["Client"])
end

function BananaDKP:OnClick()
	if hostOffline then
		self:Print(L["Will try to contact the host again now..."])
		hostOffline = nil
		self:TriggerEvent("BananaDKP_HostChanged")
		self:UpdateDisplay()
	end
end

do
	local names = {}
	local offline = "|cffff0000" .. PLAYER_OFFLINE .. "|r"
function BananaDKP:OnTooltipUpdate()
	local cat = tablet:AddCategory(
		"columns", 2,
		"child_textR", 1,
		"child_textG", 1,
		"child_textB", 0,
		"child_text2R", 1,
		"child_text2G", 1,
		"child_text2B", 1
	)
	if activeModule then
		cat:AddLine("text", L["Type"], "text2", activeModule.name)
	end
	cat:AddLine("text", L["Role"], "text2", db.mode == 1 and L["Server"] or L["Client"])
	if db.mode == 2 then
		if db.host then
			cat:AddLine("text", L["Host"], "text2", hostOffline and offline or coloredName[db.host])
		end
		cat:AddLine("text", L["Updated"], "text2", lastDKPUpdate and date("%H:%M", lastDKPUpdate) or L["Never"])
	elseif #knownClients > 0 then
		cat:AddLine("text", L["Clients"], "text2", table.concat(knownClients, ", "))
	end

	for k in pairs(names) do names[k] = nil end

	if UnitInRaid("player") then
		for i = 1, GetNumRaidMembers() do table.insert(names, (GetRaidRosterInfo(i))) end
	elseif GetNumPartyMembers() > 0 then
		table.insert(names, playerName)
		for i = 1, GetNumPartyMembers() do table.insert(names, (UnitName("party" .. i))) end
	else
		for k in pairs(guildMemberList) do
			local d = activeModule:GetDKP(k)
			if d then table.insert(names, k) end
		end
	end

	if #names > 0 then
		cat = tablet:AddCategory(
			"columns", 2,
			"text", L["Player"],
			"text2", L["DKP"],
			"child_textR", 1,
			"child_textG", 1,
			"child_textB", 0,
			"child_text2R", 1,
			"child_text2G", 1,
			"child_text2B", 1
		)
		table.sort(names, dkpSorter)
		for i, v in ipairs(names) do
			cat:AddLine("text", coloredName[v], "text2", activeModule:GetDKP(v) or 0)
		end
	end
end
end

-----------------------------------------------------------------------
-- DKP Module Handling
-----------------------------------------------------------------------

function BananaDKP:GetMode()
	return db.mode
end

function BananaDKP:GetHost()
	return hostOffline and nil or db.host
end

local bidHelp = L["After the loot master says \"bid: [Item]\", you can type \"bid\" to express your interest in the item, the winner is picked automatically."]
local dkpHelp = L["Whisper the loot master \"bdkp\" to get your current DKP, or whisper \"bdkp <name>\" to check the DKP of someone else."]
local function helpCommand(event, topic)
	if topic then
		if whisperHelp[topic] then
			return whisperHelp[topic]
		elseif topic == "bid" then
			return bidHelp
		elseif topic == "dkp" then
			return dkpHelp
		end
	else
		local help = {}
		table.insert(help, bidHelp)
		table.insert(help, dkpHelp)
		for k, v in pairs(whisperHelp) do
			table.insert(help, v)
		end
		return help
	end
end

function BananaDKP:SetActiveDKPModule(moduleName)
	assert(not running, "Can't change the DKP type while active.")
	if not self:HasModule(moduleName) then return end
	db.dkpType = moduleName

	-- Clear the registered whisper handlers.
	for k in pairs(whisperHandlers) do whisperHandlers[k] = nil end
	for k in pairs(whisperHelp) do whisperHelp[k] = nil end

	self:RegisterWhisperHandler("help", helpCommand, L["The help command gives you instructions on how to use other commands."])

	local mod = self:GetModule(moduleName)
	for name, module in self:IterateModules() do
		if name == moduleName then
			self:ToggleModuleActive(mod, true)
			options.args.dkpTypeOptions.handler = mod
			options.args.dkpTypeOptions.args = mod.options
		else
			self:ToggleModuleActive(module, false)
		end
	end

	activeModule = mod
end

function BananaDKP:RegisterWhisperHandler(prefix, handlerFunc, prefixHelp)
	assert(type(prefix) == "string", "Invalid prefix.")
	assert(not whisperHandlers[prefix], "Prefix already registered.")

	whisperHelp[prefix] = prefixHelp
	whisperHandlers[prefix] = handlerFunc
end

function BananaDKP:RegisterDKPType(name, module)
	if module:IsRegistered() then
		error(name .. " has already been registered.")
	end
	if type(module.defaultDatabase) ~= "table" then
		error(name .. " needs a default database.")
	end
	if type(module.defaultDatabase.databaseVersion) ~= "number" then
		error(name .. "'s database needs a databaseVersion number field.")
	end
	if type(module.GetDKP) ~= "function" then
		error(name .. " needs a GetDKP method.")
	end

	module.db = self:AcquireDBNamespace(name)
	self:RegisterDefaults(name, "profile", module.defaultDatabase)

	self:ToggleModuleActive(module, false)

	local opt = options.args.dkpType.args

	opt[name] = {
		type = "toggle",
		name = name,
		desc = L["Activate %s."]:format(name),
		get = function() return BananaDKP.db.profile.dkpType == name end,
		set = "SetActiveDKPModule",
		passValue = name,
		disabled = function() return running end,
	}

	if type(module.OnRegister) == "function" then
		module:OnRegister()
	end
end

-----------------------------------------------------------------------
-- DKP Type prototype
-----------------------------------------------------------------------

function BananaDKP.modulePrototype:OnInitialize()
	assert(not self.registered, self.name .. " has already been registered.")
	BananaDKP:RegisterDKPType(self.name, self)
end

function BananaDKP.modulePrototype:IsRegistered()
	return self.registered
end

-----------------------------------------------------------------------
-- Sitout list handling
-----------------------------------------------------------------------

function BananaDKP:ProcessSitoutWhisper(player)
	if not guildMemberList[player] then
		self:Print(L["%q tried to add himself to our sitout list, but he's not in our guild."]:format(player))
		return
	end
	if db.sitouts[player] then
		self:RemoveSitoutPlayer(player)
		self:Chat(L["BDKP: You have been REMOVED from the sitout list. Whisper me again to add yourself."], "WHISPER", nil, player)
	else
		self:AddSitoutPlayer(player)
		self:Chat(L["BDKP: You have been ADDED to the sitout list. Whisper me again to remove yourself."], "WHISPER", nil, player)
	end
end

function BananaDKP:ClearSitoutList()
	for k in pairs(db.sitouts) do db.sitouts[k] = nil end
	self:RebuildSitoutList()
end

function BananaDKP:AddSitoutPlayer(name)
	db.sitouts[name] = true
	self:Print(L["%q added to the sitout list."]:format(name))
	self:RebuildSitoutList()
end

function BananaDKP:RemoveSitoutPlayer(name)
	db.sitouts[name] = nil
	self:Print(L["%q removed from the sitout list."]:format(name))
	self:RebuildSitoutList()
end

function BananaDKP:RebuildSitoutList()
	local opts = options.args.sitouts.args

	for k, v in pairs(opts) do
		if k ~= "add" and k ~= "clear" and k ~= "spacer" then
			opts[k].hidden = true
		end
	end

	for k, v in pairs(db.sitouts) do
		if not opts[k] then
			opts[k] = {
				type = "execute",
				name = coloredName[k],
				desc = L["Remove %s."]:format(k),
				func = "RemoveSitoutPlayer",
				passValue = k,
				order = 100,
			}
		end
		opts[k].hidden = nil
	end
end

